"""Marker panel management for different Molecular Pixelation assays.

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

from pathlib import Path

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


from typing import TYPE_CHECKING, Optional

import pandas as pd

from pixelator.common.config import AntibodyPanel
from pixelator.common.types import PathType
from pixelator.common.utils import logger

if TYPE_CHECKING:
    from pixelator.config.config_class import AntibodyPanelMetadata
    from pixelator.pna.config.config_class import PNAConfig


class PNAAntibodyPanel(AntibodyPanel):
    """Class representing a Molecular Pixelation antibody panel."""

    # required columns
    _REQUIRED_COLUMNS = [
        "marker_id",
        "control",
        "nuclear",
        "sequence_1",
        "sequence_2",
        "conj_id",
    ]

    # and these should have unique values
    _UNIQUE_COLUMNS = ["sequence_1", "sequence_2", "conj_id"]

    def __init__(
        self,
        df: pd.DataFrame,
        metadata: AntibodyPanelMetadata,
        file_name: Optional[str] = None,
    ) -> None:
        """Load a panel from a dataframe and metadata.

        :param df: The dataframe containing the panel information.
        :param metadata: The metadata for the panel.
        :param file_name: The optional basename of the file from which
            the panel is loaded.

        :returns: None
        :raises AssertionError: exception if panel file is missing,
                                invalid or with incorrect format
        """
        self._filename = file_name
        self._metadata = metadata
        self._df = df

        # validate the panel
        errors = self.validate_antibody_panel(df)
        if len(errors) > 0:
            msg_str = "\n".join(errors)
            raise AssertionError(
                f"The following errors were found validating the panel: {msg_str}"
            )

    @classmethod
    def from_csv(cls, filename: PathType) -> Self:
        """Create an AntibodyPanel from a csv panel file.

        :param filename: The path to the panel file.
        :returns: The AntibodyPanel object.
        :raises AssertionError: exception if panel file is missing,
        :rtype: AntibodyPanel
        """
        panel_file = Path(filename)

        if not panel_file.is_file() or panel_file.suffix != ".csv":
            raise AssertionError(
                f"Panel file {filename} not found or has an incorrect format"
            )

        logger.debug("Creating Antibody panel from file %s", filename)

        df = cls._parse_panel(panel_file)
        metadata = cls._parse_header(panel_file)

        logger.debug("Antibody panel from file %s created", filename)

        return cls(df, metadata, file_name=panel_file.name)

    @classmethod
    def _parse_panel(cls, panel_file: Path) -> pd.DataFrame:
        panel = pd.read_csv(str(panel_file), comment="#")

        # validate the panel
        errors = cls.validate_antibody_panel(panel)
        if len(errors) > 0:
            msg_str = "\n".join(errors)
            raise AssertionError(
                f"The following errors were found validating the panel: {msg_str}"
            )

        # assign the sequence (unique) as index
        panel.index = panel.conj_id  # type: ignore

        # return a local copy
        return panel.copy()

    @classmethod
    def validate_antibody_panel(cls, panel_df: pd.DataFrame) -> list[str]:
        """Validate the antibody panel dataframe.

        :param panel_df: The dataframe containing the panel information.
        :returns: A list of errors found in the panel.
        """
        errors = super().validate_antibody_panel(panel_df)

        if any(panel_df["marker_id"].str.contains("_")):
            # Markers should not contain underscores since this messes
            # things up with Seurat on the R side
            errors.append(
                "The marker_id column should not contain underscores. "
                "Please use dashes instead. Offending values: "
                f"{panel_df['marker_id'][panel_df['marker_id'].str.contains('_')]}"
            )

        return errors


def load_antibody_panel(config: PNAConfig, panel: PathType) -> PNAAntibodyPanel:
    """Load an antibody panel from a file or from the config file.

    :param config: the config object
    :param panel: the path to the panel file or the name of the
        panel in the config file

    :returns: the antibody panel
    :rtype: PNAAntibodyPanel
    """
    panel_str = str(panel)
    panel_from_config = config.get_panel(panel_str)

    if panel_from_config is not None:
        logger.info("Found panel in config file: %s", panel_from_config.name)
        return panel_from_config

    panel_obj = PNAAntibodyPanel.from_csv(panel)
    return panel_obj



"""Pipeline for processing amplicon reads, code adopted from cutadapt.

Modifications Copyright © 2025 Pixelgen Technologies AB.
Under the same license terms as the original code.

Original copyright notice:

Copyright (c) 2010 Marcel Martin <marcel.martin@scilifelab.se> and contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
import typing
from typing import Iterable, Optional, Tuple

from cutadapt.files import InputFiles
from cutadapt.info import ModificationInfo
from cutadapt.modifiers import (
    PairedEndModifier,
    PairedEndModifierWrapper,
    SingleEndModifier,
)
from cutadapt.pipeline import Pipeline
from cutadapt.steps import PairedEndStep, SingleEndStep
from cutadapt.utils import Progress

from pixelator.pna.read_processing.modifiers import CombiningModifier

PairedModifiers = (
    PairedEndModifier | tuple[SingleEndModifier | None, SingleEndModifier | None]
)


ModifierStage = typing.Literal["pre", "post"]


class AmpliconPipeline(Pipeline):
    """Processing pipeline for combining paired-end reads into a single amplicon sequence.

    A series of modifiers and steps can be applied to the reads before and after the combining step.
    The pipeline is designed to work with paired-end reads only.

    :param combiner: The step that combines the reads into a single amplicon sequence.
    :param pre_modifiers: A list of modifiers that are applied to the reads before the combining step.
    :param pre_steps: A list of steps that are applied to the reads before the combining step.
    :param post_modifiers: A list of modifiers that are applied to the reads after the combining step.
    :param post_steps: A list of steps that are applied to the reads after the combining step.
    """

    paired = True

    def __init__(
        self,
        combiner: CombiningModifier,
        pre_modifiers: Iterable[PairedEndModifier] | None = None,
        pre_steps: Iterable[PairedEndStep] | None = None,
        post_modifiers: Iterable[SingleEndModifier] | None = None,
        post_steps: Iterable[SingleEndStep] | None = None,
    ):
        """Create a new AmpliconPipeline instance.

        :param combiner: The step that combines the reads into a single amplicon sequence.
        :param pre_modifiers: A list of modifiers that are applied to the reads before the combining step.
        :param pre_steps: A list of steps that are applied to the reads before the combining step.
        :param post_modifiers: A list of modifiers that are applied to the reads after the combining step.
        :param post_steps: A list of steps that are applied to the reads after the combining step.
        """
        self._combiner = combiner
        self._pre_modifiers: list[PairedEndModifier] = []
        self._pre_steps: list[PairedEndStep] = list(pre_steps) if pre_steps else []
        self._post_modifiers: list[SingleEndModifier] = []
        self._post_steps: list[SingleEndStep] = list(post_steps) if post_steps else []

        self._reader = None

        # Whether to ignore pair_filter mode for discard-untrimmed filter
        self.override_untrimmed_pair_filter = False

        # Load the pre- and post-modifiers from the constructor arguments
        if pre_modifiers:
            self._add_modifiers(pre_modifiers, stage="pre")
        if post_modifiers:
            self._add_modifiers(post_modifiers, stage="post")

    @property
    def _modifiers(self) -> list[PairedEndModifier]:
        """Return a list of all modifiers in the pipeline."""
        return self._pre_modifiers + self._post_modifiers

    @property
    def _steps(self) -> list[PairedEndModifier]:
        """Return a list of all steps in the pipeline."""
        return self._pre_steps or [] + [self._combiner] + self._post_steps or []

    def _add_modifiers(self, modifiers, stage: ModifierStage):
        for modifier in modifiers:
            if isinstance(modifier, tuple):
                self._add_two_single_modifiers(modifier[0], modifier[1], stage=stage)
            else:
                self._add_modifier(modifier, stage=stage)

    def _add_two_single_modifiers(
        self,
        modifier1: Optional[SingleEndModifier],
        modifier2: Optional[SingleEndModifier],
        stage: ModifierStage,
    ) -> None:
        """Add two single-end modifiers that modify R1 and R2, respectively.

        One of them can be None, in which case the modifier
        is only applied to the respective other read.

        :param modifier1: The modifier for read 1.
        :param modifier2: The modifier for read 2.
        :param stage: The stage in which to apply the modifiers
        """
        assert stage in ("pre", "post")

        if modifier1 is None and modifier2 is None:
            raise ValueError("Not both modifiers can be None")

        if stage == "pre":
            self._pre_modifiers.append(PairedEndModifierWrapper(modifier1, modifier2))
        if stage == "post":
            self._post_modifiers.append(PairedEndModifierWrapper(modifier1, modifier2))

    def _add_modifier(self, modifier: PairedEndModifier, stage: ModifierStage) -> None:
        """Add a Modifier (without wrapping it in a PairedEndModifierWrapper)."""
        assert stage in ("pre", "post")

        if stage == "pre":
            self._pre_modifiers.append(modifier)
        if stage == "post":
            self._post_modifiers.append(modifier)

    def process_reads(
        self,
        infiles: InputFiles,
        progress: Optional[Progress] = None,
    ) -> Tuple[int, int, Optional[int]]:
        """Receive a slice of reads and process them through the pipeline.

        :param infiles: A list of input files to process.
        """
        self._infiles = infiles
        self._reader = infiles.open()
        n = 0  # no. of processed reads
        total1_bp = 0
        total2_bp = 0
        assert self._reader is not None

        pre_modifiers_and_steps = self._pre_modifiers + self._pre_steps
        post_modifiers_and_steps = self._post_modifiers + self._post_steps

        # Note that this reader is not a real file but a chunk backed by a BytesIO object
        # in case the pipeline is running on the ParallelPipelineRunner.
        # Each worker will have its own chunk of reads to process in its own pipeline instance.
        for reads in self._reader:
            n += 1
            if n % 10000 == 0 and progress is not None:
                progress.update(10000)
            read1, read2 = reads
            total1_bp += len(read1)
            total2_bp += len(read2)
            info1 = ModificationInfo(read1)
            info2 = ModificationInfo(read2)

            for step in pre_modifiers_and_steps:
                reads = step(*reads, info1, info2)  # type: ignore
                if reads is None:
                    break

            if reads is not None:
                read1, read2 = reads
                combined_read = self._combiner(read1, read2, info1, info2)
                combined_read_info = ModificationInfo(combined_read)

                if combined_read is not None:
                    for step in post_modifiers_and_steps:
                        try:
                            read = step(combined_read, combined_read_info)  # type: ignore
                        except Exception as e:
                            logging.error(
                                "Error in step %s for read %s", step, combined_read.name
                            )
                            raise

                        if read is None:
                            break

        if progress is not None:
            progress.update(n % 10000)

        infiles.close()
        return (n, total1_bp, total2_bp)



"""Copyright © 2024 Pixelgen Technologies AB."""

import logging

import pandas as pd
import polars as pl
from anndata import AnnData

from pixelator.pna.config import PNAAntibodyPanel

logger = logging.getLogger(__name__)


def calculate_antibody_metrics(counts_df):
    """Calculate antibody metrics from a count matrix, can be used to set/update adata.vars."""
    total_antibody = pd.Series(counts_df.sum(axis=0), name="antibody_count")
    relative_antibody = pd.Series(
        total_antibody / total_antibody.sum(), name="antibody_pct"
    )
    components_detected = pd.Series((counts_df > 0).sum(axis=0), name="components")

    return pd.concat([total_antibody, relative_antibody, components_detected], axis=1)


def add_panel_information(adata, panel):
    """Add extra panel variables to var."""
    panel_meta_data = panel.df[["marker_id", "control", "nuclear"]]
    panel_meta_data = panel_meta_data.set_index("marker_id")
    panel_meta_data.index = panel_meta_data.index.astype(str)
    panel_meta_data = panel_meta_data.fillna("no")

    adata.var = adata.var.join(panel_meta_data, how="left")
    return adata


def pna_edgelist_to_anndata(edgelist: pl.LazyFrame, panel: PNAAntibodyPanel) -> AnnData:
    """Build an AnnData object from a PNA edgelist and a panel object."""

    def construct_marker_count_matrix(edgelist):
        marker_1_counts = (
            edgelist.unique(["umi1", "marker_1"])
            .select(["marker_1", "component"])
            .group_by(["component", "marker_1"])
            .agg(pl.len().alias("marker_1_count"))
            .with_columns(marker="marker_1")
            .select("component", "marker", "marker_1_count")
        )

        marker_2_counts = (
            edgelist.unique(["umi2", "marker_2"])
            .select(["marker_2", "component"])
            .group_by(["component", "marker_2"])
            .agg(pl.len().alias("marker_2_count"))
            .with_columns(marker="marker_2")
            .select("component", "marker", "marker_2_count")
        )
        marker_count = (
            marker_1_counts.join(
                marker_2_counts, on=["component", "marker"], how="full"
            )
            .with_columns(
                marker_fixed=pl.when(pl.col("marker").is_null())
                .then(pl.col("marker_right"))
                .otherwise(pl.col("marker")),
                component_fixed=pl.when(pl.col("component").is_null())
                .then(pl.col("component_right"))
                .otherwise(pl.col("component")),
                count=pl.col("marker_1_count").fill_null(0)
                + pl.col("marker_2_count").fill_null(0),
            )
            .select(
                [
                    pl.col("component_fixed").alias("component"),
                    pl.col("marker_fixed").alias("marker"),
                    pl.col("count"),
                ]
            )
            .collect()
            .pivot(on="marker", index="component", values="count")
            .fill_null(0)
        )
        return marker_count

    def component_metrics(edgelist, counts_df):
        grouped_by_component = edgelist.group_by("component")
        a_markers = grouped_by_component.agg(pl.col("umi1").n_unique().alias("n_umi1"))
        b_markers = grouped_by_component.agg(pl.col("umi2").n_unique().alias("n_umi2"))
        edges = grouped_by_component.agg(pl.len().alias("n_edges"))
        reads_in_component = grouped_by_component.agg(
            pl.col("read_count").sum().alias("reads_in_component")
        )
        info_agg = (
            a_markers.join(b_markers, on="component")
            .join(edges, on="component")
            .join(reads_in_component, on="component")
            .collect()
            .to_pandas()
            .set_index("component")
            .astype("uint64")
        )
        markers = pd.DataFrame(
            pd.Series((counts_df > 0).sum(axis=1), name="n_antibodies", dtype="uint64")
        )
        df = pd.concat([info_agg, markers], axis=1)
        df["n_umi"] = df["n_umi1"] + df["n_umi2"]

        return df

    logger.debug("Constructing counts matrix.")
    counts_df = construct_marker_count_matrix(edgelist).to_pandas()
    counts_df = counts_df.set_index("component")
    counts_df = counts_df.reindex(columns=panel.markers, fill_value=0)
    counts_df.columns = counts_df.columns.astype(str)

    # compute components metrics (obs) and re-index
    logger.debug("Computing component metrics.")
    components_metrics_df = component_metrics(edgelist, counts_df)
    components_metrics_df = components_metrics_df.reindex(index=counts_df.index)
    components_metrics_df.index = components_metrics_df.index.astype(str)

    # compute antibody metrics (var) and re-index
    logger.debug("Computing antibody metrics.")
    antibody_metrics_df = calculate_antibody_metrics(counts_df=counts_df)
    antibody_metrics_df = antibody_metrics_df.reindex(index=panel.markers, fill_value=0)
    # Do a dtype conversion of the columns here since AnnData cannot handle
    # a pyarrow arrays.
    antibody_metrics_df = antibody_metrics_df.astype(
        {"antibody_count": "int64", "antibody_pct": "float32"}
    )
    antibody_metrics_df.index = antibody_metrics_df.index.astype(str)

    # create AnnData object
    counts_df.index = counts_df.index.astype(
        str
    )  # anndata requires indexes to be strings
    logger.debug("Building AnnData instance.")
    adata = AnnData(
        X=counts_df,
        obs=components_metrics_df,
        var=antibody_metrics_df,
    )

    adata = add_panel_information(adata, panel)

    # find fraction of isotype markers in cell
    total_marker_counts = counts_df.sum(axis=1)
    isotype_markers = adata.var[adata.var["control"] == "yes"].index
    isotype_counts = counts_df[isotype_markers].sum(axis=1)
    adata.obs["isotype_fraction"] = isotype_counts / total_marker_counts

    intracellular_markers = adata.var[adata.var["nuclear"] == "yes"].index
    adata.obs["intracellular_fraction"] = (
        counts_df[intracellular_markers].sum(axis=1) / total_marker_counts
    )

    return adata



"""Generate layouts for components in a Pixel dataset.

Copyright © 2024 Pixelgen Technologies AB.
"""

import tempfile
from pathlib import Path
from typing import Iterable

import polars as pl

from pixelator.common.graph.backends.protocol import SupportedLayoutAlgorithm
from pixelator.pna.analysis_engine import (
    PerComponentTask,
)
from pixelator.pna.graph import PNAGraph
from pixelator.pna.pixeldataset.io import PixelFileWriter, PxlFile


class CreateLayout(PerComponentTask):
    """Run one or more layout algorithms on a component."""

    TASK_NAME = "layout"

    def __init__(
        self,
        layout_algorithms: list[SupportedLayoutAlgorithm],
        algorithm_kwargs: dict | None = None,
    ) -> None:
        """Create a new CreateLayout instance.

        :param layout_algorithms: The layout algorithms to run.
        :param algorithm_kwargs: Additional keyword arguments to pass to the layout algorithms.
        """
        self._layout_algorithms = layout_algorithms
        self._algorithm_kwargs = algorithm_kwargs or {}

    def run_on_component_graph(
        self, component: PNAGraph, component_id: str
    ) -> pl.LazyFrame:
        """Run the layout on a component.

        :param component: The component to run the analysis on. Either a Graph or a LazyFrame.
        :param component_id: The id of the component.
        :return: a LazyFrame containing the layout data.
        :raises TypeError: If the component is not a Graph or a LazyFrame.
        """
        results = []
        for algo in self._layout_algorithms:
            # TODO to get things working working with setting a different
            # k for pmgs w need to pass the weights from here, or change
            # the code in pixelator to have that as a parameter
            layout = pl.DataFrame(
                component.layout_coordinates(
                    algo,
                    get_node_marker_matrix=False,
                    **self._algorithm_kwargs,
                )
            )
            layout = layout.with_columns(
                component=pl.lit(component_id),
                graph_projection=pl.lit("full"),
                layout=pl.lit(algo),
            )
            results.append(layout)

        concatenated = pl.concat(results, how="vertical").lazy()
        return concatenated

    def run_on_component_edgelist(
        self, component: pl.LazyFrame, component_id: str
    ) -> pl.LazyFrame:
        """Run the layout on a component.

        :param component: The component to run the analysis on. Either a Graph or a LazyFrame.
        :param component_id: The id of the component.
        :return: a LazyFrame containing the layout data.
        """
        raise NotImplementedError

    def concatenate_data(self, data: Iterable[pl.LazyFrame]) -> pl.LazyFrame:
        """Concatenate the data. Override this if you need custom concatenation behavior."""
        return pl.concat(data, how="vertical_relaxed")

    def add_to_pixel_file(self, data: pl.LazyFrame, pxl_file_target: PxlFile) -> None:
        """Add the data in the right place in the pxl_dataset."""
        with tempfile.NamedTemporaryFile(suffix=".parquet") as tmp_file:
            tmp_file = Path(tmp_file.name)  # type: ignore
            data.sink_parquet(tmp_file)  # type: ignore
            with PixelFileWriter(pxl_file_target.path) as writer:
                writer.write_layouts(tmp_file)
